/* Taken from Etherboot and simplified to always use TSC and long longs,
 * and eliminate the use of the good old 18.2Hz ticks. Ticks are now 1000
 * times per second. --ts1 */

/* A couple of routines to implement a low-overhead timer for drivers */

 /*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 */

#include	<lib.h>
#include	<arch/io.h>
#include	<arch/timer.h>
#define DEBUG_THIS DEBUG_TIMER
#include <debug.h>

/* Ports for the 8254 timer chip */
#define	TIMER2_PORT	0x42
#define	TIMER_MODE_PORT	0x43

/* Meaning of the mode bits */
#define	TIMER0_SEL	0x00
#define	TIMER1_SEL	0x40
#define	TIMER2_SEL	0x80
#define	READBACK_SEL	0xC0

#define	LATCH_COUNT	0x00
#define	LOBYTE_ACCESS	0x10
#define	HIBYTE_ACCESS	0x20
#define	WORD_ACCESS	0x30

#define	MODE0		0x00
#define	MODE1		0x02
#define	MODE2		0x04
#define	MODE3		0x06
#define	MODE4		0x08
#define	MODE5		0x0A

#define	BINARY_COUNT	0x00
#define	BCD_COUNT	0x01

/* Timers tick over at this rate */
#define CLOCK_TICK_RATE	1193180U
#define	TICKS_PER_MS	(CLOCK_TICK_RATE/1000)

/* Parallel Peripheral Controller Port B */
#define	PPC_PORTB	0x61

/* Meaning of the port bits */
#define	PPCB_T2OUT	0x20	/* Bit 5 */
#define	PPCB_SPKR	0x02	/* Bit 1 */
#define	PPCB_T2GATE	0x01	/* Bit 0 */

static void load_timer2(unsigned int ticks)
{
	/*
	 * Now let's take care of PPC channel 2
	 *
	 * Set the Gate high, program PPC channel 2 for mode 0,
	 * (interrupt on terminal count mode), binary count,
	 * load 5 * LATCH count, (LSB and MSB) to begin countdown.
	 *
	 * Note some implementations have a bug where the high bits byte
	 * of channel 2 is ignored.
	 */
	/* Set up the timer gate, turn off the speaker */
	/* Set the Gate high, disable speaker */
	outb((inb(PPC_PORTB) & ~PPCB_SPKR) | PPCB_T2GATE, PPC_PORTB);
	/* binary, mode 0, LSB/MSB, Ch 2 */
	outb(TIMER2_SEL|WORD_ACCESS|MODE0|BINARY_COUNT, TIMER_MODE_PORT);
	/* LSB of ticks */
	outb(ticks & 0xFF, TIMER2_PORT);
	/* MSB of ticks */
	outb(ticks >> 8, TIMER2_PORT);
}

static int timer2_running(void)
{
	return ((inb(PPC_PORTB) & PPCB_T2OUT) == 0);
}

static inline unsigned long long rdtsc(void)
{
	unsigned long long val;
	 __asm__ __volatile__ ("rdtsc" : "=A" (val));
	 return val;
}

#define LATCH 0xffff

static unsigned long long calibrate_tsc(void)
{
	unsigned long long start, end;
	
	load_timer2(LATCH);
	start = rdtsc();
	while (timer2_running())
	    ;
	end = rdtsc();

	/* convert to KHz */
	return (end - start) * CLOCK_TICK_RATE / (1000*LATCH);
}

static unsigned long long cpu_khz;
void setup_timers(void)
{
	if (!cpu_khz) {
		cpu_khz = calibrate_tsc();
		/* Display the CPU MHz to easily test if the calibration was bad */
		debug("CPU %Ld MHz\n", (cpu_khz+500) / 1000);
	}
}

unsigned long currticks(void)
{
	/* Read the Time Stamp Counter */
	return  rdtsc() / cpu_khz;
}

void ndelay(unsigned int nsecs)
{
	unsigned long long timeout = rdtsc() + nsecs * cpu_khz / (1000*1000);
	while (rdtsc() < timeout)
	    ;
}

void udelay(unsigned int usecs)
{
	unsigned long long timeout = rdtsc() + usecs * cpu_khz / 1000;
	while (rdtsc() < timeout)
	    ;
}

void mdelay(unsigned int msecs)
{
	unsigned long long timeout = rdtsc() + msecs * cpu_khz;
	while (rdtsc() < timeout)
	    ;
}

void delay(unsigned int secs)
{
	unsigned long long timeout = rdtsc() + secs * cpu_khz * 1000;
	while (rdtsc() < timeout)
	    ;
}
