#include <arch/stdint.h>
#include <lib.h>

void console_init(void)
{
#ifdef VGA_CONSOLE
    vga_init();
#endif
#ifdef BTEXT_CONSOLE
    btext_init();
#endif
#ifdef SERIAL_CONSOLE
    serial_init();
#endif
}

static void device_putchar(unsigned char c)
{
#ifdef VGA_CONSOLE
    vga_putchar(c);
#endif
#ifdef BTEXT_CONSOLE
    btext_putchar(c);
#endif
#ifdef SERIAL_CONSOLE
    serial_putchar(c);
#endif
}

int last_putchar;

int putchar(int c)
{
    c &= 0xff;
    last_putchar = c;
    if (c == '\n')
	device_putchar('\r');
    device_putchar(c);
    return c;
}

/* Surprisingly, gcc-3.x sometimes converts calls to printf to puts or putchar! */
int puts(const char *s)
{
    int n = 0;

    while (*s) {
        putchar(*s++);
	n++;
    }
    putchar('\n');
    return n+1;
}

int havekey(void)
{
#ifdef SERIAL_CONSOLE
    if (serial_havechar())
	return 1;
#endif
#ifdef PC_KEYBOARD
    if (kbd_havekey())
	return 1;
#endif
    return 0;
}

int havechar(void)
{
#ifdef SERIAL_CONSOLE
    if (serial_havechar())
	return 1;
#endif
#ifdef PC_KEYBOARD
    if (kbd_havechar())
	return 1;
#endif
    return 0;
}

int getchar(void)
{
    for (;;) {
#ifdef SERIAL_CONSOLE
	if (serial_havechar())
	    return serial_getchar();
#endif
#ifdef PC_KEYBOARD
	if (kbd_havechar())
	    return kbd_getchar();
#endif
    }
}

int getline(char *buf, int max)
{
    int cur, ch, nonspace_seen;

    cur = 0;
    while (buf[cur]) {
	putchar(buf[cur]);
	cur++;
    }
    for (;;) {
	ch = getchar();
	switch (ch) {
	/* end of line */
	case '\r':
	case '\n':
	    putchar('\n');
	    goto out;
	/* backspace */
	case '\b':
	case '\x7f':
	    if (cur > 0) {
		cur--;
		putchar('\b');
		putchar(' ');
		putchar('\b');
	    }
	    break;
	/* word erase */
	case 'W' & 0x1f: /* ^W */
	    nonspace_seen = 0;
	    while (cur) {
		if (buf[cur-1] != ' ')
		    nonspace_seen = 1;
		putchar('\b');
		putchar(' ');
		putchar('\b');
		cur--;
		if (nonspace_seen && cur < max-1 && cur > 0 && buf[cur-1]==' ')
		    break;
	    }
	    break;
	/* line erase */
	case 'U' & 0x1f: /* ^U */
	    while (cur) {
		putchar('\b');
		putchar(' ');
		putchar('\b');
		cur--;
	    }
	    cur = 0;
	    break;
	default:
	    if (ch < 0x20)
		break; /* ignore control char */
	    if (ch >= 0x7f)
		break;
	    if (cur + 1 < max) {
		putchar(ch); /* echo back */
		buf[cur] = ch;
		cur++;
	    }
	}
    }
out:
    if (cur >= max)
	cur = max - 1;
    buf[cur] = '\0';
    return cur;
}
