/* Simple malloc functions
 * 2003-06 by SONE Takeshi */

#include <stddef.h>
#include <lib.h>

#define DEBUG_THIS DEBUG_MALLOC
#include <debug.h>

#ifdef TEST_MALLOC
# define DEBUG
# define malloc my_malloc
# define free my_free
# define calloc my_calloc
# define realloc my_realloc
#endif

/* Assumes sizeof(long) == sizeof(void*) */

#define INUSE (1<<0)
#define STRIP_FLAG(x) ((x) & ~1)

/* Header of each block */
struct block_header {
    /* LSBs of these two variables are used for INUSE flag */
    unsigned long prev_size; /* size of previous block */
    unsigned long size; /* size of this block */
};

#define NEXT(x) ((struct block_header *) ((unsigned long) (x) + STRIP_FLAG((x)->size)))
#define PREV(x) ((struct block_header *) ((unsigned long) (x) - STRIP_FLAG((x)->prev_size)))

#define ALIGNMENT (sizeof(struct block_header))
#define ALIGN(x) (((x) + (ALIGNMENT-1)) & ~(ALIGNMENT-1))

extern char _heap, _eheap;

static unsigned long heap_start = (unsigned long) &_heap;
static unsigned long heap_end = (unsigned long) &_eheap;

static struct block_header *head, *tail, *free_space;

static unsigned long total_free, total_alloc;
static unsigned int count_free, count_alloc;

static void malloc_init(void)
{
    head = (struct block_header *) ALIGN(heap_start);
    head->prev_size = INUSE;
    head->size = ((heap_end & ~(ALIGNMENT-1)) - ALIGN(heap_start))
        - sizeof(*tail);

    tail = NEXT(head);
    tail->prev_size = head->size;
    tail->size = INUSE;

    free_space = head;

    total_free = head->size;
    total_alloc = 0;
    count_free = 1;
    count_alloc = 0;
#ifdef DEBUG
    malloc_check();
    malloc_diag();
#endif
}

void *malloc(size_t size)
{
    struct block_header *p, *new_space;
    unsigned long required, largest, remainder;

    if (!head)
	malloc_init();

#ifdef DEBUG
    malloc_check();
#endif

    required = sizeof(struct block_header) + ALIGN(size);

    if (!free_space || free_space->size < required) {
	largest = 0;
	new_space = 0;
	for (p = head; p < tail; p = NEXT(p)) {
	    if ((p->size & INUSE) == 0 && p->size > largest) {
		new_space = p;
		largest = p->size;
	    }
	}
	free_space = new_space;
	if (largest < required) {
	    printf("Out of heap space\n");
#ifdef TEST_MALLOC
	    return 0;
#else
	    abort();
#endif
	}
    }

    remainder = free_space->size - required;
    p = free_space;
    p->size = required | INUSE;
    NEXT(p)->prev_size = p->size;
    if (remainder) {
	free_space = NEXT(p);
	free_space->size = remainder;
        NEXT(free_space)->prev_size = free_space->size;
    } else {
	free_space = 0;
	count_free--;
    }

    total_free -= required;
    total_alloc += required;
    count_alloc++;

#ifdef DEBUG
    malloc_check();
    malloc_diag();
#endif
    return p+1;
}

void free(void *mem)
{
    struct block_header *p, *next, *prev;

#ifdef DEBUG
    malloc_check();
#endif

    if (!mem) {
	printf("Attempt to free NULL\n");
        abort();
    }

    p = ((struct block_header *) mem) - 1;
    next = NEXT(p);

    /* sanity check */
    if ((p->size & INUSE) == 0 || (NEXT(p)->prev_size != p->size)) {
	printf("Attempt to free non-allocated memory: %p\n", mem);
        abort();
    }

    p->size &= ~INUSE;
    next->prev_size = p->size;

    total_free += p->size;
    total_alloc -= p->size;
    count_alloc--;
    count_free++;

    /* consolidate with next block */
    if ((next->size & INUSE) == 0) {
	p->size += next->size;
	next = NEXT(p);
	next->prev_size = p->size;
	count_free--;
    }

    /* consolidate with previous block */
    if ((p->prev_size & INUSE) == 0) {
	prev = PREV(p);
	prev->size += p->size;
	p = prev;
	next->prev_size = p->size;
	count_free--;
    }

    if (!free_space || p->size > free_space->size)
	free_space = p;

#ifdef DEBUG
    malloc_check();
    malloc_diag();
#endif
}

void malloc_diag(void)
{
    debug("alloc: %lu bytes (%u blocks), free: %lu bytes (%u blocks)\n",
	    total_alloc, count_alloc, total_free, count_free);
}

#ifdef DEBUG
void malloc_check(void)
{
    struct block_header *p;
    unsigned long t_alloc, t_free;
    unsigned int c_alloc, c_free;


    if (head->prev_size != INUSE) {
	debug("invalid head->prev_size: %#lx\n", head->prev_size);
        abort();
    }

    t_alloc = t_free = 0;
    c_alloc = c_free = 0;

    for (p = head; p < tail; p = NEXT(p)) {
	if (NEXT(p) > tail) {
	    debug("invalid size: %#lx at %p\n", p->size, p);
            abort();
	}
	if (NEXT(p)->prev_size != p->size) {
	    debug("sizes mismatch: %#lx vs %#lx at %p\n", p->size,
		    NEXT(p)->prev_size, p);
            abort();
	}
	if ((p->size & INUSE) == 0) {
	    if ((p->prev_size & INUSE) == 0)
		debug("successive free blocks: %p\n", p);
	    t_free += p->size;
	    c_free++;
	} else {
	    t_alloc += STRIP_FLAG(p->size);
	    c_alloc++;
	}
    }

    if (t_free != total_free) {
	debug("total_free incorrect: %lu vs %lu\n", total_free, t_free);
        abort();
    }

    if (c_free != count_free) {
	debug("count_free incorrect: %u vs %u\n", count_free, c_free);
        abort();
    }

    if (t_alloc != total_alloc) {
	debug("total_alloc incorrect: %lu vs %lu\n", total_alloc, t_alloc);
        abort();
    }

    if (c_alloc != count_alloc) {
	debug("count_alloc incorrect: %u vs %u\n", count_alloc, c_alloc);
        abort();
    }
}
#endif /* DEBUG */

void *calloc(size_t nmemb, size_t size)
{
    size_t alloc_size = nmemb * size;
    void *mem;

    if (alloc_size < nmemb || alloc_size < size) {
	printf("calloc overflow: %u, %u\n", nmemb, size);
	return 0;
    }

    mem = malloc(alloc_size);
    memset(mem, 0, alloc_size);

    return mem;
}

void *realloc(void *mem, size_t size)
{
    size_t copy_size;
    struct block_header *h;
    void *new_mem;

    if (mem == 0)
	return malloc(size);
    if (size == 0) {
	free(mem);
	return 0;
    }

    h = ((struct block_header *) mem) - 1;
    copy_size = h->size;
    if (size < copy_size)
	copy_size = size;
    /* XXX should optimze this */
    new_mem = malloc(size);
    memcpy(new_mem, mem, copy_size);
    free(mem);
    return new_mem;
}

#ifdef TEST_MALLOC
#include <stdio.h>
//#include <stdlib.h>

#define NBLOCKS 8192

char heap[65536];
void *blocks[NBLOCKS];

int main()
{
    size_t max_size;
    int nblocks, i, pass, r;
    void *mem, *tmp;

    heap_start = (unsigned long) heap;
    heap_end = heap_start + sizeof heap;
    malloc_init();

    pass = 0;
    nblocks = 0;
    for (;;) {
        max_size = sizeof heap;
        for (; nblocks < NBLOCKS; nblocks++) {
            for (;;) {
                mem = malloc(rand() % max_size);
                if (mem)
                    break;
                max_size >>= 1;
                if (max_size == 0)
                    goto filled;
            }
            printf("malloc %p\n", mem);
            blocks[nblocks] = mem;
        }
    filled:
        for (i = 0; i < nblocks; i++) {
            r = rand() % nblocks;
            tmp = blocks[i];
            blocks[i] = blocks[r];
            blocks[r] = tmp;
        }

        for (i = nblocks/2; i < nblocks; i++) {
            //printf("free %p\n", blocks[i]);
            free(blocks[i]);
        }
        nblocks /= 2;
        printf("PASS %d\n", ++pass);
    }
}

#endif /* TEST_MALLOC */
